from netmiko import BaseConnection
from napalm.base import models as napalm_models, NetworkDriver

from ..base import SCBaseSSHDriver


class PSIMN(SCBaseSSHDriver, NetworkDriver):
    """
        Very sketchy, basic driver for noc-wan-nokia-psi-dc
        running 1830PSIMN-25.6-0 to pull config backups.

    Netmiko doesn't appear to model this so we're
    doing a paramiko-based driver instead.
    """

    def open(self):
        """
        Implementation of NAPALM method 'open' to open a connection to the device.
        """
        SCBaseSSHDriver.open(self)

        # switching to a different CLI that we know how to run commands
        self.device.send_command(
            "//\n", expect_string="Switching to MD-CLI", cmd_verify=False
        )

    def get_config(
        self, retrieve="all", full=False, sanitized=False, format="text"
    ) -> napalm_models.ConfigDict:
        config = self.send_command("admin show running | no-more")

        return {"startup": {}, "candidate": {}, "running": config}



from netmiko import BaseConnection
from napalm.base import models as napalm_models, NetworkDriver

from ..base import SCBaseSSHDriver


class CienaRLS(SCBaseSSHDriver, NetworkDriver):
    """
    Netmiko-based Ciena RLS SSH driver
    """

    def get_config(
        self, retrieve="all", full=False, sanitized=False, format="text"
    ) -> napalm_models.ConfigDict:
        config = self.send_command("show-config", read_timeout=60)

        return {"startup": {}, "candidate": {}, "running": config}



from typing import Optional, Dict, List, Union, Any
import socket
import re

from napalm.base import NetworkDriver
from napalm.base.helpers import textfsm_extractor
from napalm.base.netmiko_helpers import netmiko_args
from napalm.base.exceptions import ConnectionClosedException
import napalm.base.models as napalm_models


from ..base import SCBaseDriver
from ..models import InventoryDict
from ..utils import age_to_integer


class SCIOSXR_SSH(SCBaseDriver, NetworkDriver):
    """
    Netmiko-based napalm class for pulling data from an IOS-XR device.
    """

    PROTOCOL_ABBRS = {
        "L": "local",
        "C": "connected",
        "S": "static",
        "i": "ISIS",
        "B": "BGP",
        "O": "OSPF",
    }

    I_ABBRS = {
        "Lo": "Loopback",
        "BE": "Bundle-Ether",
        "Hu": "HundredGigE",
        "Te": "TenGigE",
        "Mg": "MgmtEth",
        "FH": "FourHundredGigE",
    }

    INVENTORY_TO_TYPE = {
        r"Fan Tray": "fan",
        r"Pluggable": "optic",
        r"Line Card": "linecard",
        r"Route Processor": "re",
        r"Fabric Card": "fabric_module",
        r"System Controller": None,
        r"Chassis": None,
        r"Power Supply": "psu",
        r".*": None,  # Cannot deal with weird wan things I just really want optics
    }

    def __init__(
        self,
        hostname: str,
        username: str,
        password: str,
        timeout: int = 60,
        optional_args: Optional[Dict] = None,
    ) -> None:
        self.hostname = hostname
        self.username = username
        self.password = password
        self.timeout = timeout

        self.device = None

        optional_args = optional_args if optional_args else {}
        optional_args["read_timeout_override"] = timeout
        self.netmiko_optional_args = netmiko_args(optional_args)

        self._i_descrs = None

    @property
    def i_descrs(self):
        """
        Pulls interface descriptions into an internal data structure
        """

        if self._i_descrs is None:
            raw_output = self._send_command("show interface description")
            output = textfsm_extractor(self, "sh_interface_description", raw_output)
            self._i_descrs = {o["interface"]: o["description"] for o in output}

        return self._i_descrs

    def _send_command(self, command):
        """
        This is a copy of "_send_command" from the napalm ios driver
        """
        try:
            if isinstance(command, list):
                for cmd in command:
                    output = self.device.send_command(cmd)
                    if "% Invalid" not in output:
                        break
            else:
                output = self.device.send_command(command)
            return output
        except (socket.error, EOFError) as e:
            raise ConnectionClosedException(str(e))

    def open(self):
        """
        Netmiko open
        """
        device_type = "cisco_xr"
        self.device = self._netmiko_open(
            device_type, netmiko_optional_args=self.netmiko_optional_args
        )

    def close(self):
        """
        Netmiko close
        """
        self._netmiko_close()

    def cli(
        self, commands: List[str], encoding="text"
    ) -> Dict[str, Union[str, Dict[str, Any]]]:
        """
        Executes an arbitrary list of commands and returns as a dict
        mapping the executed command to its result.
        """
        return {cmd: self._send_command(cmd) for cmd in commands}

    def get_facts(self) -> napalm_models.FactsDict:
        """
        Processes "show version" into napalm facts dict
        """

        facts = {
            "os_version": "",
            "uptime": 0,
            "interface_list": self.i_descrs.keys(),
            "vendor": "cisco",
            "serial_number": "",
            "model": "",
            "hostname": "",
            "fqdn": "",
        }

        # show version has os version, hostname, and uptime
        show_version = self._send_command("show version detail")
        for line in show_version.split("\n"):
            m = re.match(r"Cisco IOS XR Software, Version (\S+)$", line)
            if m:
                facts["os_version"] = m.group(1)
                continue

            m = re.match(r"(\S+) System uptime is (.+)$", line)
            if m:
                facts["hostname"] = m.group(1)
                facts["uptime"] = age_to_integer(m.group(2))
                continue

        # show inventory chassis has model and serial number
        show_inv_chassis = self._send_command("admin show inventory chassis")
        for line in show_inv_chassis.split("\n"):
            m = re.match(r"\s+PID: (\S+).+SN: (\S+)", line)
            if m:
                facts["model"] = m.group(1)
                facts["serial_number"] = m.group(2)

        return facts

    def get_config(
        self,
        retrieve: str = "all",
        full: bool = False,
        sanitized: bool = False,
        format: str = "text",
    ) -> napalm_models.ConfigDict:
        """retrieves configuration"""

        if sanitized:
            raise NotImplementedError("Sanitized config not supported")

        if format != "text":
            raise NotImplementedError("Only text config format supported")

        # running config and startup config are the same
        results = {"running": "", "startup": "", "candidate": ""}
        if retrieve in ["all", "running", "startup"]:
            config = self._send_command("show running-config")

            if retrieve in ["all", "running"]:
                results["running"] = config
            if retrieve in ["all", "startup"]:
                results["startup"] = config

        # candidate config is only valid if we're in config mode
        if retrieve in ["all", "candidate"] and self.device.check_config_mode():
            results["candidate"] = self._send_command("show configuration merge")

        return results

    def get_lldp_neighbors(self) -> Dict[str, List[napalm_models.LLDPNeighborDict]]:
        """
        Parses show lldp neighbors - doing "show lldp neighbors detail" beacuse otherwise
        the hostname gets cut off
        """
        raw_output = self._send_command("show lldp neighbors detail")
        output = textfsm_extractor(self, "sh_lldp_neighbors_detail", raw_output)

        neighs = {}
        for neigh in output:
            if neigh["local_port"] not in neighs:
                neighs[neigh["local_port"]] = []

            neighs[neigh["local_port"]].append(
                {"hostname": neigh["name"], "port": neigh["port"]}
            )
        return neighs

    def get_lldp_neighbors_detail(
        self, interface: str = ""
    ) -> napalm_models.LLDPNeighborsDetailDict:
        """ "
        Parses show lldp neighbors detail
        """

        raw_output = self._send_command("show lldp neighbors detail")
        output = textfsm_extractor(self, "sh_lldp_neighbors_detail", raw_output)

        neighs = {}
        for neigh in output:
            if interface and neigh["local_port"] != interface:
                continue

            if neigh["local_port"] not in neighs:
                neighs[neigh["local_port"]] = []

                neighs[neigh["local_port"]].append(
                    {
                        "parent_interface": "",
                        "remote_chassis_id": neigh["chassis_id"],
                        "remote_system_name": neigh["name"],
                        "remote_port": neigh["port"],
                        "remote_port_description": neigh["port_descr"],
                        "remote_system_description": neigh["descr"],
                        "remote_system_capab": neigh["capab"].split(","),
                        "remote_system_enable_capab": neigh["enable_capab"].split(","),
                    }
                )
        return neighs

    def get_arp_table(self) -> List[napalm_models.ARPTableDict]:
        """
        Parses "show arp ; show arp vrf all"
        """
        raw_output = self._send_command("show arp ; show arp vrf all")
        output = textfsm_extractor(self, "sh_arp", raw_output)

        entries = []
        for entry in output:
            entries.append(
                {
                    "interface": entry["interface"],
                    "mac": entry["mac"],
                    "ip": entry["ip"],
                    "age": entry["age"] if entry["age"] != "-" else "",
                }
            )

        return entries

    def get_inventory(self) -> List[InventoryDict]:
        """
        Parses "show invetory"
        """
        raw_output = self._send_command("show inventory")
        output = textfsm_extractor(self, "sh_inventory", raw_output)

        inventory = []
        for entry in output:
            inv_type = self._get_inventory_type(entry["desc"])
            if not inv_type:
                continue

            inventory.append(
                {
                    "name": entry["name"],
                    "type": inv_type,
                    "part_number": entry["pid"],
                    "serial_number": entry["sn"],
                }
            )

        return inventory

    def get_optics(self):
        """
        ssh based "get optics". Just like other platforms,
        For this platform you need to get levels on each
        individual interface
        """
        interfaces = self._send_command("show interfaces brief")
        phys = textfsm_extractor(self, "sh_int_brief_phy", interfaces)

        output = {}

        for phy in phys:
            i_name = f"{phy['prefix']}{phy['slot']}"
            output[i_name] = {"physical_channels": {"channels": {}}}
            i_channels = []

            ctrlr = self._send_command(f"show controllers optics {phy['slot']}")
            levels = textfsm_extractor(self, "controllers_optics", ctrlr)

            for channel in levels:
                i_channels.append(
                    {
                        "index": channel["lane"] if channel["lane"] else 0,
                        "input_power": channel["rx"],
                        "output_power": channel["tx"],
                        "laser_bias_current": channel["bias"],
                    }
                )
            output[i_name]["physical_channels"]["channels"] = i_channels

        return output



from typing import List, Literal, Dict
from napalm.eos import EOSDriver
from napalm.base import models as napalm_models
from netaddr import EUI, mac_unix_expanded
import re

from ..base import SCBaseDriver
from ..models import InventoryDict

from pprint import pprint


class SCEOSDriver(EOSDriver, SCBaseDriver):
    INVENTORY_TO_TYPE = {
        r"Fabric": "fabric_module",
        r"Linecard": "linecard",
        r"Supervisor": "re",
    }

    SC_BACKUP_CMDS = [
        "wr mem",
        "show running-config sanitized",
        "show version detail",
        "show interface status",
        "show inventory",
        "show int hardware",
        "show isis neighbors",
        "show bgp summary",
        "show logging",
        "show lldp neighbor",
    ]
    SC_SHOW_TECH_CMDS = ["show tech"]

    def __init__(self, hostname, username, password, timeout=60, optional_args=None):
        """
        Forcing ssh transport since we don't enable the web interface
        """
        optional_args = optional_args if optional_args else {}
        optional_args["transport"] = "ssh"

        super().__init__(
            hostname, username, password, timeout=timeout, optional_args=optional_args
        )
        self.netmiko_optional_args["read_timeout_override"] = 600

    def get_sc_backup(self) -> Dict[str, Literal["cmds", "show_tech"]]:
        results = self._run_commands(self.SC_BACKUP_CMDS, encoding="text")

        output = {}
        output["cmds"] = ""
        for cmd, cmd_output in zip(self.SC_BACKUP_CMDS, results):
            output["cmds"] += f"########## {cmd} ############\n"
            output["cmds"] += cmd_output["output"] + "\n"

        show_tech = self._run_commands(self.SC_SHOW_TECH_CMDS, encoding="text")
        output["show_tech"] = show_tech[0]["output"]

        return output

    def get_inventory(self) -> List[InventoryDict]:
        output = self._run_commands(
            ["show inventory", "show interface transceiver"], encoding="json"
        )
        inventory = output[0]

        # 'show int transciever' gives us a mapping of the slot number that shows up
        # in 'show inventory' to the name of the interface.
        trans = output[1]["interfaces"]

        slot_to_port = {}
        for i_name, item in trans.items():
            if not item:
                continue
            slot = item["slot"].replace("Ethernet", "")

            if slot not in slot_to_port:
                slot_to_port[slot] = i_name
        results = []

        ### optics
        for slot, optic in inventory["xcvrSlots"].items():
            if slot not in slot_to_port:
                continue

            if optic.get("serialNum"):
                results.append(
                    {
                        "type": "optic",
                        "subtype": optic["modelName"],
                        "name": slot_to_port[slot],
                        "part_number": optic["modelName"],
                        "serial_number": optic["serialNum"],
                    },
                )

        ### line cards
        for slot, card in inventory["cardSlots"].items():
            if card.get("serialNum"):
                results.append(
                    {
                        "type": self._get_inventory_type(slot),
                        "subtype": card["modelName"],
                        "name": f"Ethernet{slot}",
                        "part_number": card["modelName"],
                        "serial_number": card["serialNum"],
                    },
                )

        ### PSUs
        for slot, psu in inventory["powerSupplySlots"].items():
            if psu.get("serialNum"):
                results.append(
                    {
                        "type": "psu",
                        "subtype": None,
                        "name": f"PSU {slot}",
                        "part_number": psu["name"],
                        "serial_number": psu["serialNum"],
                    },
                )

        ### FANs
        for slot, fan in inventory["fanTraySlots"].items():
            if fan.get("serialNum"):
                results.append(
                    {
                        "type": "fan",
                        "subtype": None,
                        "name": f"FAN {slot}",
                        "part_number": fan["name"],
                        "serial_number": fan["serialNum"],
                    },
                )
        return results

    def get_ipv6_neighbors_table(self) -> List[napalm_models.IPV6NeighborDict]:
        """
        Not implemented in Napalm base EOS!
        """
        output = []
        neighbors = self._run_commands(["show ipv6 neighbors vrf all"], encoding="json")
        if not neighbors:
            return output

        for _, vrf in neighbors[0]["vrfs"].items():
            for entry in vrf["ipV6Neighbors"]:
                mac = EUI(entry["hwAddress"])
                mac.dialect = mac_unix_expanded
                output.append(
                    {
                        "interface": entry["interface"],
                        "mac": str(mac),
                        "ip": entry["address"],
                        "age": entry["age"],
                        "state": "REACH",
                    }
                )
        return output

    def get_optics(self):
        """
        napalm eos driver is problematic in several ways so we get to rewrite it
        """
        command = ["show interfaces transceiver"]

        output = self._run_commands(command, encoding="json")[0]["interfaces"]

        # Formatting data into return data structure
        optics_detail = {}

        for port, port_values in output.items():
            if not port_values:
                continue

            # multi-channel optics show up as separate ports, eg
            # Ethernet45/1, Ethernet45/2, Ethernet45/3. We're going to
            # use the 'slot' attribute instead
            port_name = port_values["slot"] + "/1"
            if port_name not in optics_detail:
                optics_detail[port_name] = {"physical_channels": {"channel": []}}

            # Defaulting avg, min, max values to 0.0 since device does not
            # return these values
            optic_states = {
                "index": port_values["channel"],
                "state": {
                    "input_power": {
                        "instant": (
                            port_values["rxPower"] if "rxPower" in port_values else 0.0
                        ),
                        "avg": 0.0,
                        "min": 0.0,
                        "max": 0.0,
                    },
                    "output_power": {
                        "instant": (
                            port_values["txPower"] if "txPower" in port_values else 0.0
                        ),
                        "avg": 0.0,
                        "min": 0.0,
                        "max": 0.0,
                    },
                    "laser_bias_current": {
                        "instant": (
                            port_values["txBias"] if "txBias" in port_values else 0.0
                        ),
                        "avg": 0.0,
                        "min": 0.0,
                        "max": 0.0,
                    },
                },
            }

            optics_detail[port_name]["physical_channels"]["channel"].append(
                optic_states
            )

        return optics_detail



from typing import List
import logging

from napalm.nxos_ssh import NXOSSSHDriver
from napalm.base.helpers import textfsm_extractor

from ..base import SCBaseDriver
from ..models import (
    InventoryDict,
)

logger = logging.getLogger(__name__)


class SCNXOS(NXOSSSHDriver, SCBaseDriver):
    """
    NXOS Parser
    """

    # for nexus we're going to map the 'description' provided by
    # show inventory to the type
    INVENTORY_TO_TYPE = {
        # note we're using the fact that this dict gets evaluated
        # sequentially to catch the linecards, whose descriptions are varied
        # but all end in 'Module'
        r"Fabric Module": "fabric_module",
        r"Fabric card": "fabric_module",  # N7K fabric modules in umd
        r"Fabric Extender": None,  # FEXes in Dearborn
        r"N2K-C2": "stack_member",  # FEXes in Dearborn
        r"Eth\s?Mod": None,  # chassis for fixed config Nexus
        r"Supervisor Module": "re",
        r"Fan Module": "fan",
        r"Module": "linecard",
        r"System Controller": None,
        r"Chassis": None,
        r"Power Supply": "psu",
    }

    I_ABBRS = {
        "Lo": "loopback",
        "Po": "port-channel",
        "Eth": "Ethernet",
    }

    def get_inventory(self) -> List[InventoryDict]:
        """
        Parses "show inventory" and "show interface transciever"
        """

        raw_inventory = self._send_command("show inventory")
        inventory = textfsm_extractor(self, "sh_inventory", raw_inventory)

        output = []
        for entry in inventory:
            inventory_type = self._get_inventory_type(entry["desc"])
            if not inventory_type:
                continue

            output.append(
                {
                    "type": inventory_type,
                    "name": entry["name"],
                    "part_number": entry["pid"],
                    "serial_number": entry["sn"],
                }
            )

        raw_trans = self._send_command("show interface transceiver")
        trans = textfsm_extractor(self, "sh_int_transceiver", raw_trans)
        for entry in trans:
            if "AOC" in entry["type"]:
                db_type = "aoc"
            elif "DAC" in entry["type"]:
                db_type = "dac"
            else:
                db_type = "optic"

            output.append(
                {
                    "type": db_type,
                    "subtype": entry["type"],
                    "name": entry["interface"],
                    "part_number": entry["pn"],
                    "serial_number": entry["sn"],
                }
            )

        return output



NS = {
    "dn-top": "http://drivenets.com/ns/yang/dn-top",
    "dn-sys": "http://drivenets.com/ns/yang/dn-system",
    "dn-sys-dns": "http://drivenets.com/ns/yang/dn-sys-dns",
    "dn-sys-ncp": "http://drivenets.com/ns/yang/dn-sys-ncp",
    "dn-platform": "http://drivenets.com/ns/yang/dn-platform",
    "dn-if": "http://drivenets.com/ns/yang/dn-interfaces",
    "dn-trans": "http://drivenets.com/ns/yang/dn-transceivers",
    "dn-lldp": "http://drivenets.com/ns/yang/dn-lldp",
    "dn-proto": "http://drivenets.com/ns/yang/dn-protocol",
}

FACTS_RPC_REQ = """<get xmlns="urn:ietf:params:xml:ns:netconf:base:1.1">
  <filter>
    <drivenets-top xmlns="http://drivenets.com/ns/yang/dn-top">
      <system xmlns="http://drivenets.com/ns/yang/dn-system">
        <oper-items>
          <system-type/>
          <system-uptime/>
          <system-version/>
        </oper-items>
        <config-items>
          <name/>
        </config-items>
        <ncps xmlns="http://drivenets.com/ns/yang/dn-sys-ncp">
          <ncp>
            <config-items>
              <platform>
                <oper-items>
                  <serial-number/>
                </oper-items>
              </platform>
            </config-items>
          </ncp>
        </ncps>
        <dns xmlns="http://drivenets.com/ns/yang/dn-sys-dns">
          <config-items>
            <domain-name/>
          </config-items>
        </dns>
      </system>
      <interfaces xmlns="http://drivenets.com/ns/yang/dn-interfaces">
        <interface>
          <name/>
        </interface>
      </interfaces>
    </drivenets-top>
  </filter>
</get>
"""

OPTICS_RPC_REQ = """<get xmlns="urn:ietf:params:xml:ns:netconf:base:1.1">
<filter>
  <drivenets-top xmlns="http://drivenets.com/ns/yang/dn-top">
    <interfaces xmlns="http://drivenets.com/ns/yang/dn-interfaces">
       <interface>
          <transceivers xmlns="http://drivenets.com/ns/yang/dn-transceivers">
            <oper-items>
              <digital-optical-monitoring/>
            </oper-items>
        </transceivers>
      </interface>
    </interfaces>
  </drivenets-top>
</filter>
</get>
"""

LLDP_NEIGH_RPC_REQ = """<get xmlns="urn:ietf:params:xml:ns:netconf:base:1.1">
<filter>
	<drivenets-top xmlns="http://drivenets.com/ns/yang/dn-top">
    <protocols xmlns="http://drivenets.com/ns/yang/dn-protocol">
      <lldp xmlns="http://drivenets.com/ns/yang/dn-lldp">
        <interfaces>
          <interface>
            <neighbors>
              <neighbor/>
            </neighbors>
          </interface>
        </interfaces>
      </lldp>      
    </protocols>
	</drivenets-top>
</filter>
</get>
"""

INVENTORY_RPC_REQ = """<get xmlns="urn:ietf:params:xml:ns:netconf:base:1.1">
<filter>
  <drivenets-top xmlns="http://drivenets.com/ns/yang/dn-top">
    <interfaces xmlns="http://drivenets.com/ns/yang/dn-interfaces">
       <interface>
        <oper-items>
          <interface-speed/>
        </oper-items>
        <transceivers xmlns="http://drivenets.com/ns/yang/dn-transceivers">
          <oper-items>
            <form-factor/>
            <ethernet-pmd/>
            <vendor/>
            <vendor-part/>
            <serial-no/>
            <wavelength/>
          </oper-items>
        </transceivers>
      </interface>
    </interfaces>
  </drivenets-top>
</filter>
</get>
"""



NS = {
    "int": "http://cisco.com/ns/yang/Cisco-IOS-XR-pfi-im-cmd-oper",
    "int4": "http://cisco.com/ns/yang/Cisco-IOS-XR-ipv4-io-oper",
    "int6": "http://cisco.com/ns/yang/Cisco-IOS-XR-ipv6-ma-oper",
    "rib4": "http://cisco.com/ns/yang/Cisco-IOS-XR-ip-rib-ipv4-oper",
    "rib6": "http://cisco.com/ns/yang/Cisco-IOS-XR-ip-rib-ipv6-oper",
    "inv": "http://cisco.com/ns/yang/Cisco-IOS-XR-invmgr-oper",
    "opt": "http://cisco.com/ns/yang/Cisco-IOS-XR-controller-optics-oper",
}


IP_INT_RPC_REQ = """
<get xmlns="urn:ietf:params:xml:ns:netconf:base:1.1">
  <filter>
    <interfaces xmlns="http://cisco.com/ns/yang/Cisco-IOS-XR-pfi-im-cmd-oper">
      <interfaces>
        <interface/>
      </interfaces>
    </interfaces>
    <ipv4-network xmlns="http://cisco.com/ns/yang/Cisco-IOS-XR-ipv4-io-oper">
      <nodes>
        <node>
          <interface-data>
            <vrfs>
              <vrf>
                <details>
                  <detail/>
                </details>
              </vrf>
            </vrfs>
          </interface-data>
        </node>
      </nodes>
    </ipv4-network>
    <ipv6-network xmlns="http://cisco.com/ns/yang/Cisco-IOS-XR-ipv6-ma-oper">
      <nodes>
        <node>
          <interface-data>
            <vrfs>
              <vrf>
                <global-details>
                  <global-detail/>
                </global-details>
              </vrf>
            </vrfs>
          </interface-data>
        </node>
      </nodes>
    </ipv6-network>
  </filter>
</get>"""

IP_ROUTE_RPC_REQ = """
<get xmlns="urn:ietf:params:xml:ns:netconf:base:1.1">
  <filter>
    <rib xmlns="http://cisco.com/ns/yang/Cisco-IOS-XR-ip-rib-ipv4-oper">
      <vrfs>
        <vrf>
          <afs>
            <af>
              <af-name>IPv4</af-name>
              <safs>
                <saf>
                  <saf-name>Unicast</saf-name>
                  <ip-rib-route-table-names>
                    <ip-rib-route-table-name>
                      <route-table-name>default</route-table-name>
                      <routes>
                        <route>
                          <prefix-length/>
                          <protocol-name/>
                          <route-age/>
                          <route-path>
                            <active>true</active>
                          </route-path>
                        </route>
                      </routes>
                    </ip-rib-route-table-name>
                  </ip-rib-route-table-names>
                </saf>
              </safs>
            </af>
          </afs>
        </vrf>
      </vrfs>
    </rib>
  
    <ipv6-rib xmlns="http://cisco.com/ns/yang/Cisco-IOS-XR-ip-rib-ipv6-oper">
      <vrfs>
        <vrf>
          <afs>
            <af>
              <af-name>IPv6</af-name>
              <safs>
                <saf>
                  <saf-name>Unicast</saf-name>
                  <ip-rib-route-table-names>
                    <ip-rib-route-table-name>
                      <route-table-name>default</route-table-name>
                      <routes>
                        <route>
                          <prefix-length/>
                          <protocol-name/>
                          <route-age/>
                          <route-path>
                            <active>true</active>
                          </route-path>
                        </route>
                      </routes>
                    </ip-rib-route-table-name>
                  </ip-rib-route-table-names>
                </saf>
              </safs>
            </af>
          </afs>
        </vrf>
      </vrfs>
    </ipv6-rib>

  </filter>
</get>
"""

INV_RPC_REQ = """
<get xmlns="urn:ietf:params:xml:ns:netconf:base:1.1">
  <filter>
    <inventory xmlns="http://cisco.com/ns/yang/Cisco-IOS-XR-invmgr-oper">
      <entities>
        <entity>
          <attributes>
            <inv-basic-bag>
              <description/>
              <name/>
              <serial-number/>
              <manufacturer-name/>
              <model-name/>
              <is-field-replaceable-unit>true</is-field-replaceable-unit>
            </inv-basic-bag>
          </attributes>
        </entity>
      </entities>
    </inventory>
  </filter>
</get>
"""

OPTICS_RPC_REQ = """
<get xmlns="urn:ietf:params:xml:ns:netconf:base:1.1">
  <filter>
    <optics-oper xmlns="http://cisco.com/ns/yang/Cisco-IOS-XR-controller-optics-oper">
      <optics-ports>
        <optics-port>
          <optics-lanes>
            <optics-lane>
              <number/>
              <transmit-power/>
              <receive-power/>
              <laser-bias-current-milli-amps/>
            </optics-lane>
          </optics-lanes>
        </optics-port>
      </optics-ports>
    </optics-oper>
  </filter>
</get>
"""



from napalm_srl import NokiaSRLDriver
from napalm_srl.srl import SRLAPI
from napalm.base import models as napalm_models
from ..base import SCBaseNetconfDriver

from ipaddress import IPv6Address

from napalm.base.helpers import convert

import logging


class SCSLRAPI(SRLAPI):
    """
    This override fixes an issue with IPv6 addresses embedded in URLs when talking
    over GRPCs.
    """

    def __init__(self, hostname, username, password, timeout=60, optional_args=None):
        super().__init__(hostname, username, password, timeout, optional_args)

        try:
            IPv6Address(hostname)
            self.target = f"[{hostname}]:{self.gnmi_port}"
        except ValueError:
            pass


class SCNokiaSRLDriver(NokiaSRLDriver, SCBaseNetconfDriver):
    SC_BACKUP_CMDS = [
        "save startup",
        "info",
        "show version",
        "show interface brief",
        "show platform chassis",
        "show network-instance protocols isis adjacency",
        "show network-instance protocols bgp summary",
        "show system logging file messages",
        "show lldp neighbor",
    ]

    def __init__(self, hostname, username, password, timeout=60, optional_args=None):
        """
        Forcing insecure connection for testing purposes
        """
        optional_args = optional_args if optional_args else {}
        optional_args["insecure"] = True
        optional_args["skip_verify"] = True

        optional_args["encoding"] = "JSON_IETF"

        super().__init__(
            hostname, username, password, timeout=timeout, optional_args=optional_args
        )

        self.device = SCSLRAPI(
            hostname, username, password, timeout=timeout, optional_args=optional_args
        )

    def get_interfaces(self):
        """
        Nokia driver returns interface speeds at Gbps and not Mbps which is
        wrong.
        """
        results = super().get_interfaces()
        for i_name in results.keys():
            if results[i_name]["speed"] != -1:
                results[i_name]["speed"] = 1000 * results[i_name]["speed"]

        return results

    def get_config(
        self, retrieve="all", full=False, sanitized=False, format="text"
    ) -> napalm_models.ConfigDict:
        """
        pulling running config via ssh sow we can have it in 'display set' format.
        """
        if format != "text":
            return super().get_config(
                retrieve=retrieve, full=full, sanitized=sanitized, format=format
            )

        config = {"startup": "", "running": "", "candidate": ""}

        with self.ssh_conn() as ssh_device:
            if retrieve in ["all", "running"]:
                config["running"] = ssh_device.send_command("info from running flat")
            if retrieve in ["all", "startup"]:
                config["startup"] = ssh_device.send_command("info from startup flat")
            if retrieve in ["all", "candidate"]:
                config["candidate"] = ssh_device.send_command(
                    "info from candidate flat"
                )

        return config

    def get_optics(self):
        path = {"/interface"}
        path_type = "STATE"
        output = self.device._gnmiGet("", path, path_type)
        interfaces = self._getObj(
            output, *["srl_nokia-interfaces:interface"], default=[]
        )
        channel_data = {}

        for i in interfaces:
            if not self._getObj(i, *["transceiver", "channel"], default=False):
                continue

            name = self._getObj(i, *["name"])
            channel = self._getObj(i, *["transceiver", "channel"], default={})[0]
            channel_data.update(
                {
                    name: {
                        "physical_channels": {
                            "channel": [
                                {
                                    "index": self._getObj(
                                        channel, *["index"], default=-1
                                    ),
                                    "state": {
                                        "input_power": {
                                            "instant": self._getObj(
                                                channel,
                                                *["input-power", "latest-value"],
                                                default=-1.0,
                                            ),
                                            "avg": -1.0,
                                            "min": -1.0,
                                            "max": -1.0,
                                        },
                                        "output_power": {
                                            "instant": self._getObj(
                                                channel,
                                                *["output-power", "latest-value"],
                                                default=-1.0,
                                            ),
                                            "avg": -1.0,
                                            "min": -1.0,
                                            "max": -1.0,
                                        },
                                        "laser_bias_current": {
                                            "instant": self._getObj(
                                                channel,
                                                *["laser-bias-current", "latest-value"],
                                                default=-1.0,
                                            ),
                                            "avg": -1.0,
                                            "min": -1.0,
                                            "max": -1.0,
                                        },
                                    },
                                }
                            ]
                        }
                    }
                }
            )

        return channel_data

    def get_inventory(self):
        path = {"/interface/transceiver"}
        path_type = "STATE"
        output = self.device._gnmiGet("", path, path_type)
        transceivers = self._getObj(
            output, *["srl_nokia-interfaces:interface"], default=[]
        )
        result = []
        for t_if in transceivers:
            trans = t_if["transceiver"]

            if "serial-number" not in trans:
                continue

            result.append(
                {
                    "type": "optic",
                    "subtype": trans["ethernet-pmd"],
                    "name": t_if["name"],
                    "part_number": trans["vendor-part-number"],
                    "serial_number": trans["serial-number"],
                }
            )

        return result

    def get_arp_table(self, vrf=""):
        """
        Copied from the community driver, which has a bug in
        the 'vrf_path' statements
        """

        try:
            arp_table = []
            subinterface_names = []

            def _find_neighbors(is_ipv4, ip_dict):
                ip_dict = eval(ip_dict.replace("'", '"'))
                neighbor_list = self._find_txt(ip_dict, "neighbor")
                if neighbor_list:
                    neighbor_list = list(eval(neighbor_list))
                    for neighbor in neighbor_list:
                        ipv4_address = ""
                        ipv6_address = ""
                        timeout = -1.0
                        reachable_time = -1.0
                        if is_ipv4:
                            ipv4_address = self._find_txt(neighbor, "ipv4-address")
                            timeout = convert(
                                float, self._find_txt(ip_dict, "timeout"), default=-1.0
                            )
                        else:
                            ipv6_address = self._find_txt(neighbor, "ipv6-address")
                            reachable_time = convert(
                                float,
                                self._find_txt(ip_dict, "reachable-time"),
                                default=-1.0,
                            )
                        arp_table.append(
                            {
                                "interface": sub_interface_name,
                                "mac": self._find_txt(neighbor, "link-layer-address"),
                                "ip": ipv4_address if is_ipv4 else ipv6_address,
                                "age": timeout if is_ipv4 else reachable_time,
                            }
                        )

            if vrf:
                vrf_path = {"network-instance[name={}]/interface".format(vrf)}
            else:
                vrf_path = {"network-instance[name=*]/interface"}
            pathType = "STATE"

            vrf_output = self.device._gnmiGet("", vrf_path, pathType)
            if not vrf_output:
                return []
            for vrf in vrf_output["srl_nokia-network-instance:network-instance"]:
                if "interface" in vrf.keys():
                    subinterface_list = self._find_txt(vrf, "interface")
                    subinterface_list = list(eval(subinterface_list))
                    for dictionary in subinterface_list:
                        if "name" in dictionary.keys():
                            subinterface_names.append(
                                self._find_txt(dictionary, "name")
                            )

            interface_path = {"interface[name=*]"}
            interface_output = self.device._gnmiGet("", interface_path, pathType)

            for interface in interface_output["srl_nokia-interfaces:interface"]:
                interface_name = self._find_txt(interface, "name")
                if interface_name:
                    sub_interface = self._find_txt(interface, "subinterface")
                    if sub_interface:
                        sub_interface = list(eval(sub_interface))
                        for dictionary in sub_interface:
                            sub_interface_name = self._find_txt(dictionary, "name")
                            if sub_interface_name in subinterface_names:
                                ipv4_data = self._find_txt(dictionary, "ipv4")
                                if ipv4_data:
                                    ipv4_data = eval(ipv4_data.replace("'", '"'))
                                    ipv4_arp_dict = self._find_txt(
                                        ipv4_data, "srl_nokia-interfaces-nbr:arp"
                                    )
                                    if ipv4_arp_dict:
                                        _find_neighbors(True, ipv4_arp_dict)

                                ipv6_data = self._find_txt(dictionary, "ipv6")
                                if ipv6_data:
                                    ipv6_data = eval(ipv6_data.replace("'", '"'))
                                    ipv6_neighbor_dict = self._find_txt(
                                        ipv6_data,
                                        "srl_nokia-if-ip-nbr:neighbor-discovery",
                                    )
                                    if ipv6_neighbor_dict:
                                        _find_neighbors(False, ipv6_neighbor_dict)
            return arp_table
        except Exception as e:
            logging.error("Error occurred : {}".format(e))



# # pylint: disable=no-member
from typing import List
import re
import logging

from napalm.junos import JunOSDriver
from napalm.base.models import LLDPNeighborDict, ConfigDict
from ..models import InventoryDict
from ..base import SCBaseNetconfDriver


from . import junos_views

logger = logging.getLogger(__name__)

XCVR_DESC_TO_INTERFACE_PREFIX = {
    r"100GBASE-": "et-",
    r"QSFP": "et-",
    r"SFP28": "et-",
    r"SFP\+": "xe-",
    r"SFP-": "ge-",
    r"CFP-": "et-",
    r"(UNKNOWN|UNSUPPORTED)": "ge-",  # default
}

DEFAULT_ARP_AGING_TIMER = 1200


class SCJunOS(JunOSDriver, SCBaseNetconfDriver):
    """
    Junos parser
    """

    netmiko_host_type = "juniper_junos"

    # mapping of inventory name regex to their netbox types
    INVENTORY_TO_TYPE = {
        r"Fan Tray": "fan",
        r"(PIC|MIC|FPC)": "linecard",
        r"Xcvr": "optic",
        r"(PEM|PSM|Power Supply)": "psu",
        r"Routing Engine": "re",
        # don't care about these inventory items
        r"(CPU|PDM|FPM|Midplane|CB|SFB|TIB)": None,
    }

    SC_BACKUP_CMDS = [
        #    "commit synchronize",
        "show configuration",
        "show configuration | display set",
        "show version",
        "show interface terse",
        "show chassis hardware",
        "show isis adjacency instance attendee",
        "show isis adjacency instance conference",
        "show isis adjacency instance dmz",
        "show isis adjacency instance exhibitor",
        "show isis adjacency instance mgmt_junos",
        "show isis adjacency instance scinet",
        "show isis adjacency instance untrust",
        "show bgp summary",
        "show log messages",
        "show lldp neighbors",
        #    "file archive source /var/log* destination /var/tmp/sc25Logs.tgz",
    ]
    SC_SHOW_TECH_CMDS = ["request support information | no-more"]

    def _get_junos_inventory_type(self, name: str, model_number: str) -> str:
        """
        Maps the name and part number of the Junos inventory item to its type
        """
        # "Xcvr" parts are always optics
        if name.startswith("Xcvr"):
            return "optic"

        # non-optic parts should always have a model number, if not
        # we don't care about them
        if not (model_number):
            return None

        # EX switches and virtual chassis save their vc members and uplink modules
        # as FPC X and PIC X - we want to classify those correctly
        if re.match(r"EX[234]", model_number):
            if re.search(r"FPC \d+", name):
                return "stack-member"

            if re.search(r"PIC \d+", name):
                return None

        # uplink modules are also saved under PIC X, must classify based
        # on their model number
        if re.match(r"EX-UM", model_number):
            return "uplink-module"

        # otherwise we want to pattern-match based on the INVENTORY_TO_TYPE
        # dictionary
        return self._get_inventory_type(name)

    def _get_xcvr_interface_prefix(self, xcvr_desc: str) -> str:
        """
        Maps xcvr description from "show chassis hardware" to
        interface prefix
        """
        for pattern, prefix in XCVR_DESC_TO_INTERFACE_PREFIX.items():
            if re.match(pattern, xcvr_desc):
                return prefix

        raise ValueError(f"{self.hostname}: Unknown xcvr type {xcvr_desc}")

    def _get_mod_number(self, mod_name: str) -> str:
        """
        Strips out "FPC|MIC|PIC" from a module name seen in "show chassis hardware"
        """
        return re.sub(r"(MIC|PIC|FPC) ", "", mod_name)

    def _get_inventory_part_number(self, item: any) -> str:
        """
        Extracts the part number from an inventory item
        which, depending on the specific item, is stashed in
        the part number, model number, or description field
        """
        if item.model_number and item.model_number != "model-number":
            return item.model_number
        return item.part_number

    def _save_inventory_item(
        self, output: list, item: any, parent: str = None, grandparent: str = None
    ) -> bool:
        """
        Extracts data from a particular inventory item object.
        Returns whether we care aobut this item or not (so we know whether or not)
        to loop over its children
        """
        # skip builtin types, or parts without model numbers that aren't transcievers
        if item.part_number == "BUILTIN":
            return False

        # get inventory type based on item name and P/N
        inv_type = self._get_junos_inventory_type(item.name, item.model_number)

        if not inv_type:
            return False

        # for transcievers, change the name from Xcvr X to be the junos interface name
        # note that we expect transcievers to always be at the sub-sub module level
        # and thus to have a grandparent and parent
        m = re.search(r"Xcvr (\d+)", item.name)
        if m:
            if not (grandparent or parent):
                raise ValueError(
                    f"{self.hostname}: No MIC and PIC found for {item.name}"
                )
            prefix = self._get_xcvr_interface_prefix(item.description)

            item_name = f"{prefix}{self._get_mod_number(grandparent)}/{self._get_mod_number(parent)}/{m.group(1)}"

        # for sub-linecards (mics or pics) we want to prepend the parent to the item name
        elif parent:
            item_name = f"{parent} {item.name}"
        else:
            item_name = item.name

        output.append(
            {
                "type": inv_type,
                "name": item_name,
                "subtype": item.description,
                "part_number": self._get_inventory_part_number(item),
                "serial_number": item.serial_number,
            }
        )

        return True

    def get_config(
        self, retrieve="all", full=False, sanitized=False, format="text"
    ) -> ConfigDict:
        """
        pulling running config via ssh sow we can have it in 'display set' format.
        """
        if format != "text":
            return super().get_config(
                retrieve=retrieve, full=full, sanitized=sanitized, format=format
            )

        config = {"startup": "", "running": "", "candidate": ""}

        with self.ssh_conn() as ssh_device:
            config["running"] = ssh_device.send_command(
                "show configuration | display set"
            )

        return config

    def get_lldp_neighbors(self) -> LLDPNeighborDict:
        """
        Overrides napalm get_lldp_neighbors so that we actually get the remote
        interface ID instead of the description.

        NAPALM's get_lldp_neighbors_detail gets this for us - note that to get this detail,
        you need to execute a command for each interface just like on the cli :(
        """
        neighs = super().get_lldp_neighbors_detail()
        output = {}
        for port, neighs in neighs.items():
            output[port] = []
            for neigh in neighs:
                output[port].append(
                    {
                        "hostname": neigh["remote_system_name"]
                        if neigh["remote_system_name"]
                        else neigh["remote_chassis_id"],
                        "port": neigh["remote_port"],
                    }
                )
        return output

    def get_inventory(self) -> List[InventoryDict]:
        """
        parses the get-chassis-inventory RPC call, which maps to "show chassis hardware"
        """
        result = junos_views.junos_inventory(self.device).get()

        output = []
        for chassis in dict(result).values():
            # note that the chassis model and s/n are at this level, but
            # that doesn't count as an 'inventory item' so we're ignoring it

            # saving modules, sub-modules and sub-sub modules
            for module in dict(chassis.modules).values():
                self._save_inventory_item(output, module)

                for sub_module in dict(module.sub_modules).values():
                    self._save_inventory_item(output, sub_module, parent=module.name)

                    for sub_sub_module in dict(sub_module.sub_sub_modules).values():
                        self._save_inventory_item(
                            output,
                            sub_sub_module,
                            parent=sub_module.name,
                            grandparent=module.name,
                        )

        return output



from typing import List
import logging

from lxml import etree

from math import log

from napalm.iosxr_netconf import IOSXRNETCONFDriver
from ncclient.xml_ import to_ele

from ..base import SCBaseNetconfDriver
from ..models import InventoryDict

from .constants import INV_RPC_REQ, OPTICS_RPC_REQ, NS

logger = logging.getLogger(__name__)


class SCIOSXR(IOSXRNETCONFDriver, SCBaseNetconfDriver):
    """
    IOSXR Class
    """

    NS = NS

    INVENTORY_TO_TYPE = {
        r"PM\d+$": "psu",
        r"FT\d+$": "fan",
        r"FC\d+$": "fabric_module",
        r"RP\d$": "re",
        r"SC\d": None,  # don't care about system controller
        r"Rack 0": None,  # this is the chassis, don't care about it
        r"^(FourHundred|Hundred|Forty)GigE": "optic",
        r"\d/\d": "linecard",
    }

    def get_inventory(self) -> List[InventoryDict]:
        """
        Gets inventory data
        """
        rpc_reply = self.device.dispatch(to_ele(INV_RPC_REQ)).xml
        xml_result = etree.fromstring(rpc_reply)

        output = []
        for item in self._xpath(xml_result, "//inv:inv-basic-bag"):
            name = self._text(item, "inv:name")
            item_type = self._get_inventory_type(name)
            if item_type:
                output.append(
                    {
                        "type": item_type,
                        "name": name,
                        "part_number": self._text(item, "inv:model-name"),
                        "serial_number": self._text(item, "inv:serial-number"),
                        "parent": None,
                    }
                )

        return output

    def get_optics(self):
        rpc_reply = self.device.dispatch(to_ele(OPTICS_RPC_REQ)).xml
        xml_result = etree.fromstring(rpc_reply)

        # print(etree.tostring(xml_result, pretty_print=True, encoding="unicode"))
        output = {}
        for item in self._xpath(xml_result, ".//opt:optics-port"):
            i_channels = {"physical_channels": {"channel": []}}
            i_name = self._text(item, "opt:name").replace("Optics", "")
            for lane in self._xpath(item, ".//opt:optics-lane"):
                i_channels["physical_channels"]["channel"].append(
                    {
                        "index": self._text(lane, ".//opt:number"),
                        "state": {
                            # power given over netconf is in...dBk?
                            "input_power": float(
                                self._text(lane, ".//opt:receive-power")
                            )
                            / 100,
                            "output_power": float(
                                self._text(lane, ".//opt:transmit-power")
                            )
                            / 100,
                            "laser_bias_current": float(
                                self._text(lane, ".//opt:laser-bias-current-milli-amps")
                            ),
                        },
                    }
                )
            output[i_name] = i_channels

        return output



from napalm.base import models as napalm_models, NetworkDriver
import re
from copy import deepcopy
from ..base import SCBaseSSHDriver


class SCCienaSAOS(SCBaseSSHDriver, NetworkDriver):
    """
    Community SAOS driver is very old and wants you to have py23_compat.
    Also they don't have get_optics. Also their get_config no longer works.
    So we are writing our own.
    """

    netmiko_host_type = "ciena_saos10"
    _xcvr_ids = None  # place to cache xcvr list

    @property
    def xcvr_ids(self):
        """
        Runs 'show xcvrs' and caches them in an internal
        list so we don't have to run this command multiple
        times for multiple getters
        """
        if not self._xcvr_ids:
            self._xcvr_ids = []
            xcvrs = self.send_command("show xcvrs")
            for line in xcvrs.split("\n"):
                m = re.search(r"\| (\d+)\s+\| (enabled|disabled)", line)
                if m:
                    self._xcvr_ids.append(m.group(1))
                    continue

        return self._xcvr_ids

    def get_config(self, retrieve="all"):
        configs = {
            "startup": "",
            "running": "",
            "candidate": "",
        }

        if retrieve in ("running", "all"):
            output = self.send_command("show running config")
            configs["running"] = output
        if retrieve in ("candidate", "all"):
            output = self.send_command("show candidate config")
            configs["candidate"]

        return configs

    def get_optics(self):
        """
        Again we're in a sad place where we can't use textfsm templates
        because of the output of the data
        """

        i_channels = {"physical_channels": {"channel": []}}

        channel_template = {
            "index": None,
            "state": {
                "input_power": None,
                "output_power": None,
                "laser_bias_current": None,
            },
        }

        value_map = {
            "Tx Power": "output_power",
            "Rx Power": "input_power",
            "Tx Laser Bias": "laser_bias_current",
        }
        output = {}

        for xcvr_id in self.xcvr_ids:
            i_channels = {"physical_channels": {"channel": []}}
            channel = deepcopy(channel_template)
            next_value = None

            diags = self.send_command(f"show xcvr diagnostics xcvr {xcvr_id}")
            for line in diags.split("\n"):
                # Start of a new channel
                m = re.match(r"\|\s+Lane Number\s+\| (\d+)\s+\|", line)
                if m:
                    # if we have a previous channel entry save it
                    if channel["index"]:
                        i_channels["physical_channels"]["channel"].append(channel)
                        channel = deepcopy(channel_template)

                    channel["index"] = m.group(1)
                    continue

                # tx/rx power and bias all have distinct headers on previous rows
                # so we have to flag them to catch them on the next row
                m = re.match(r"\|\s+(Rx Power|Tx Power|Tx Laser Bias)", line)
                if m:
                    next_value = m.group(1)

                # actual value is just on a line labeled 'current'
                m = re.match(r"\|\s+Current\s+\| ([\d\.\-]+)\s+\|", line)
                if m and next_value:
                    channel["state"][value_map[next_value]] = m.group(1)
                    next_value = None

            # save the last entry
            if channel["index"]:
                i_channels["physical_channels"]["channel"].append(channel)

            output[xcvr_id] = i_channels

        return output

    def get_inventory(self):
        """
        For now just getting optics values
        """
        output = []
        entry_template = {
            "type": "optic",
            "subtype": "",
            "name": "",
            "part_number": "",
            "serial_number": "",
        }
        value_map = {
            "Vendor PN": "part_number",
            "Vendor Serial Number": "serial_number",
            "Identifier": "subtype",
        }

        for xcvr in self.xcvr_ids:
            inv = self.send_command(f"show xcvrs xcvr {xcvr}")
            entry = deepcopy(entry_template)

            entry["name"] = xcvr

            for line in inv.split("\n"):
                m = re.match(
                    r"\|\s+(Vendor PN|Vendor Serial Number|Identifier)\s+\| (\S+)", line
                )
                if m:
                    entry[value_map[m.group(1)]] = m.group(2)
                    continue

            output.append(entry)

        return output



import re
from napalm.base import NetworkDriver
from napalm.base.helpers import textfsm_extractor
from ..base import SCBaseSSHDriver

from pprint import pprint
from copy import deepcopy


class WaveServerDriver(SCBaseSSHDriver, NetworkDriver):
    INVENTORY_TO_TYPE = {
        # note we're using the fact that this dict gets evaluated
        # sequentially to catch the linecards, whose descriptions are varied
        # but all end in 'Module'
        r"Chassis": None,  # Not saving chassis in inventory, that should come from get_facts
        r"Control Processor": "re",  # best guess? Idk what these WAN things are
        r"Access Panel": None,  # I think this is a patch panel or cable mgmt?
        r"Power": "psu",  # FEXes in Dearborn
        r"Fan Module": "fan",  # chassis for fixed config Nexus
        r"EDFA": "linecard",  # when I google this it feels vaguely linecard-y
        r"SFP": "optic",
    }

    def _get_inventory_subtype(self, entry):
        """
        Inventory subtype is really just for optics models.
        The only optics in the two scimet waveservers look like:
        '400G-FR4, SMF, 2KM QSFP-DD'
        """

        if "SFP" in entry["model"]:
            return entry["model"].split(",")[0]
        return None

    def get_config(self, retrieve="all", full=False, sanitized=False, format="text"):
        """
        Waveserver appears to have one config
        """
        result = {"startup": "", "candidate": "", "running": ""}
        if retrieve in ["all", "running"]:
            result["running"] = self.send_command("configuration show")
        if retrieve in ["all", "startup"]:
            result["startup"] = self.send_command("configuration show")
        return result

    def get_inventory(self):
        """
        Waveserver inventory
        """
        raw_inv = self.send_command("chassis inventory show")
        inv = textfsm_extractor(self, "chassis_inventory_show", raw_inv)

        output = []
        for entry in inv:
            output.append(
                {
                    "type": self._get_inventory_type(entry["model"]),
                    "subtype": self._get_inventory_subtype(entry),
                    "name": entry["unit"],
                    "part_number": entry["pn"],
                    "serial_number": entry["sn"],
                },
            )
        return output

    def get_optics(self):
        """
        Need to do 'xcvr show' to get port names then can iterate over
        'xcvr show xcvr X diagnostics'
        """
        raw_xcvrs = self.send_command("xcvr show")
        xcvrs = textfsm_extractor(self, "xcvr_show", raw_xcvrs)

        output = {}
        for xcvr in xcvrs:
            i_name = xcvr["port"]
            diags_raw = self.send_command(f"xcvr show xcvr {i_name} diagnostics")

            # textfsm templates doesn't like the weird dividers between output rows
            # so we're gonna do our parsing ourselves :(
            i_channels = {"physical_channels": {"channels": []}}
            curr_i = i_channels["physical_channels"]["channels"]

            channel_template = {
                "index": None,
                "state": {
                    "input_power": None,
                    "output_power": None,
                    "laser_bias_current": None,
                },
            }
            channel = deepcopy(channel_template)
            for line in diags_raw.split("\n"):
                # Start of a new channel
                m = re.match(
                    r"\|\s+(\d+)\s+\|\s+Laser Bias\(mA\)\|\s+([\d\.]+)\s+\|", line
                )
                if m:
                    # if we have a previous channel entry
                    if channel["index"]:
                        i_channels["physical_channels"]["channels"].append(channel)
                        channel = deepcopy(channel_template)

                    channel["index"] = m.group(1)
                    channel["state"]["laser_bias_current"] = m.group(2)
                    continue

                # Coherent optic with a single channel - no bias current
                m = re.match(
                    r"\|\s+(\d+)\s+\|\s+Tx Power \(dBm\)\|\s+([\-\d\.]+)\s+\|", line
                )
                if m:
                    channel["index"] = m.group(1)
                    channel["state"]["output_power"] = m.group(2)
                    channel["state"]["laser_bias_current"] = 0.0
                    continue

                # tx power for a channel
                m = re.match(r"\|\s+\|\s+Tx Power \(dBm\)\|\s+([\-\d\.]+)\s+\|", line)
                if m:
                    channel["state"]["output_power"] = m.group(1)
                    continue

                # rx power for a channel
                m = re.match(r"\|\s+\|\s+Rx Power \(dBm\)\|\s+([\-\d\.]+)\s+\|", line)
                if m:
                    channel["state"]["input_power"] = m.group(1)
                    continue

            # save the last entry
            if channel["index"]:
                i_channels["physical_channels"]["channels"].append(channel)

            # save interface to output
            if curr_i:
                output[i_name] = i_channels

        return output



# from .iosxr_netconf import SCIOSXR
from .iosxr import SCIOSXR_SSH
from .iosxr_netconf import SCIOSXR
from .junos import SCJunOS
from .nxos import SCNXOS
from .sros import SCNokiaSROSDriver
from .srl import SCNokiaSRLDriver
from .eos import SCEOSDriver
from .nos import SCNOSDriver
from .aoscx import SCAOSCXDriver
from .waveserver import WaveServerDriver
from .psimn import PSIMN
from .rls import CienaRLS
from .saos import SCCienaSAOS

PLATFORM_MAP = {
    "iosxr": SCIOSXR_SSH,
    "iosxr_netconf": SCIOSXR,
    "nxos": SCNXOS,
    "junos": SCJunOS,
    "sros": SCNokiaSROSDriver,
    "srl": SCNokiaSRLDriver,
    "eos": SCEOSDriver,
    "nos": SCNOSDriver,
    "waveserver": WaveServerDriver,
    "aoscx": SCAOSCXDriver,
    "psimn": PSIMN,
    "rls": CienaRLS,
    "saos": SCCienaSAOS,
}


def get_network_driver(platform: str):
    """
    Returns network driver based on platform string.
    """
    for valid_platform, driver in PLATFORM_MAP.items():
        if valid_platform == platform:
            return driver

    raise NotImplementedError(f"Unsupported platform {platform}")



from typing import Union
import logging
import sys
import argparse
from decouple import config
from os import getenv

LOG_LEVELS = {
    "DEBUG": logging.DEBUG,
    "INFO": logging.INFO,
    "WARNING": logging.WARNING,
    "ERROR": logging.WARNING,
    "CRITICAL": logging.CRITICAL,
}
LOG_FORMAT = "%(asctime)-15s  %(levelname)8s %(name)s %(message)s"


def configure_logging(
    log_level: Union[int, str],
    log_globally: bool = False,
    log_file: str = None,
    log_to_console: bool = False,
):
    """
    Configures logging for the module, or globally as indicated by the input
    """

    if log_globally:
        logger = logging.getLogger()
    else:
        module_name = __name__.split(".")[0]
        logger = logging.getLogger(module_name)

    if isinstance(log_level, str):
        log_level = LOG_LEVELS[log_level.upper()]
    logger.setLevel(log_level)

    if log_file:
        file_handler = logging.handlers.RotatingFileHandler(
            log_file, maxBytes=1024 * 1024 * 10, backupCount=20
        )
        file_handler.setFormatter(logging.Formatter(LOG_FORMAT))
        logger.addHandler(file_handler)

    if log_to_console:
        stdout_handler = logging.StreamHandler(sys.stdout)
        stdout_handler.setFormatter(logging.Formatter(LOG_FORMAT))
        logger.addHandler(stdout_handler)


def get_from_args_or_env(
    cli_arg: str, parsed_args: argparse.Namespace = None, required=True
) -> str:
    """
    Pull a value from parsed arparse, if it's not there look for it
    in .env, and if it's not there, check the user's environment.
    """
    cli_arg = cli_arg.replace("-", "_")

    if getattr(parsed_args, cli_arg, False):
        return getattr(parsed_args, cli_arg)

    env_arg = cli_arg.upper()
    if config(env_arg, None):
        return config(env_arg)

    if getenv(env_arg):
        return getenv(env_arg)

    if required:
        raise ValueError(
            f"ERROR: Please provide {cli_arg} as cli input or set as {env_arg} environment variable"
        )
    return None


def age_to_integer(age: str) -> Union[int, None]:
    """
    Across platforms age strings can be:
    10y5w, 5w4d, 05d04h, 01:10:12, 3w4d 01:02:03,
    50 days 10 hours (ASA)
    """

    # convert empty string to none
    if age == "":
        return None

    # integer age is seconds - we don't need further parsing
    # if we got an integer value
    m = re.search(r"^\d+$", age)
    if m:
        return int(age)

    # Extracting digits out of different time strings
    TIME_MATCHES = [
        r"(?P<hours>\d+):(?P<minutes>\d+):(?P<seconds>\d+)",
        r"(?P<years>\d+) years?",
        r"(?P<months>\d+) months?",
        r"(?P<weeks>\d+) weeks?",
        r"(?P<days>\d+) days?",
        r"(?P<hours>\d+) hours?",
        r"(?P<minutes>\d+) minutes?",
        r"(?P<years>\d+)y",
        r"(?P<weeks>\d+)w",
        r"(?P<days>\d+)d",
        r"(?P<hours>\d+)h",
    ]
    # Multipliers for different units
    UNIT_TO_SECONDS = {
        "years": 31536000,
        "months": 2628000,
        "weeks": 604800,
        "days": 86400,
        "hours": 3600,
        "minutes": 60,
        "seconds": 1,
    }

    seconds = 0

    # search for every time match regex. If  a match is found
    # extract data, apply multiplier, and add to our seconds total
    for match in TIME_MATCHES:
        m = re.search(match, age)
        if not m:
            continue

        for unit, multiplier in UNIT_TO_SECONDS.items():
            if unit in m.groupdict():
                seconds += int(m.groupdict()[unit]) * multiplier

    return seconds
